#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import unittest
from selinonTestCase import SelinonTestCase

from selinon import SystemState
from selinon.storage import DataStorage
from selinon.config import Config


class TestStorageAccess(SelinonTestCase):
    def test_retrieve(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     Task2
        #
        class MyStorage(DataStorage):
            def __init__(self):
                pass

            def connect(self):
                # shouldn't be called
                raise NotImplementedError()

            def disconnect(self):
                # called on destruction
                pass

            def is_connected(self):
                # return True so we can test retrieve()
                return True

            def store(self, node_args, flow_name, task_name, task_id, result):
                # shouldn't be called
                raise NotImplementedError()

            def retrieve(self, task_name, task_id):
                assert(task_name == 'Task1')
                assert(task_id == task1.task_id)
                return 0xDEADBEEF

        def _cond_access(db, node_args):
            return db.get('Task1') == 0xDEADBEEF

        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': _cond_access},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)

        # Task1 has finished, we should access the database
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        Config.storage_mapping = {'Storage1': MyStorage()}
        Config.task_mapping = {'Task1': 'Storage1'}

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        system_state.update()

        self.assertIn('Task2', self.instantiated_tasks)

    @unittest.skip("store() currently not tested")
    def test_store(self):
        # store() is called transparently by SelinonTask - not possible to directly check it without running a task
        pass

    def test_connect_and_configuration(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     Task2
        #
        class MyStorage(DataStorage):
            def __init__(self):
                pass

            def connect(self):
                raise ConnectionError()

            def disconnect(self):
                # called on destruction
                pass

            def is_connected(self):
                # return False so we can test connect()
                return False

            def store(self, node_args, flow_name, task_name, task_id, result):
                # shouldn't be called
                raise NotImplementedError()

            def retrieve(self, task_name, task_id):
                # shouldn't be called
                raise NotImplementedError()

        def _cond_access(db, node_args):
            return db.get('Task1') == 0xDEADBEEF

        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': _cond_access},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)

        # Task1 has finished, we should access the database
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        Config.storage_mapping = {'Storage1': MyStorage()}
        Config.task_mapping = {'Task1': 'Storage1'}

        with self.assertRaises(ConnectionError):
            system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
            system_state.update()



#!/usr/bin/env python3

from pymongo import MongoClient
from .dataStorage import DataStorage


class MongoStorage(DataStorage):
    """
    MongoDB database Adapter
    """
    def __init__(self, db_name, collection_name, host="localhost", port=27017):
        super(MongoStorage, self).__init__()
        self.client = None
        self.collection = None
        self.db = None
        self.host = host
        self.port = port
        self.db_name = db_name
        self.collection_name = collection_name

    def is_connected(self):
        return self.client is not None

    def connect(self):
        self.client = MongoClient(self.host, self.port)
        self.db = self.client[self.db_name]
        self.collection = self.db[self.collection_name]

    def disconnect(self):
        if self.is_connected():
            self.client.close()
            self.client = None
            self.db = None
            self.collection = None

    def retrieve(self, task_name, task_id):
        assert(self.is_connected())

        filtering = {'_id': 0}
        cursor = self.collection.find({'task_id': task_id}, filtering)

        if len(cursor) > 0:
            raise ValueError("Multiple records with same task_id found")
        elif len(cursor) == 0:
            raise FileNotFoundError("Record not found in database")

        record = cursor[0]

        assert(task_name == record['task_name'])
        return record.get('result')

    def store(self, node_args, flow_name, task_name, task_id, result):
        assert(self.is_connected())

        record = {
            'node_args': node_args,
            'flow_name': flow_name,
            'task_name': task_name,
            'task_id': task_id,
            'result': result

        }

        self.collection.insert(record)

        # task_id is unique here
        return task_id




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from selinonTestCase import SelinonTestCase

from selinon import FlowError
from selinon import SystemState


class TestNodeFailures(SelinonTestCase):
    def test_single_failure_flow_fail(self):
        #
        # flow1:
        #
        #     Task1 X
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {}, 'fallback': []}}
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        with self.assertRaises(FlowError):
            system_state = SystemState(id(self), 'flow1', state=state_dict,
                                       node_args=system_state.node_args)
            system_state.update()

        self.assertNotIn('Task2', self.instantiated_tasks)

    def test_single_failure_flow_fallback_true(self):
        #
        # flow1:
        #
        #     Task1 X .... True
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {}, 'fallback': True}}

        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

    def test_single_failure_flow_fallback_start(self):
        #
        # flow1:
        #
        #     Task1 X .... Task3
        #       |            |
        #       |            |
        #     Task2        Task4
        #
        # Note:
        #   Task1 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task4'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {}, 'fallback': ['Task3']}}

        }
        self.init(edge_table, failures=failures)

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        task3 = self.get_task('Task3')
        self.assertEqual(task3.node_args, node_args)

    def test_single_failure_flow_fallback_start2(self):
        #
        # flow1:
        #
        #     Task1 X .... Task3, Task4
        #       |
        #       |
        #     Task2
        #
        #     Task3      Task4
        #       |          |
        #       |          |
        #     Task5      Task6
        #
        # Note:
        #   Task1 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task5'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task6'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {}, 'fallback': ['Task3', 'Task4']}}

        }
        self.init(edge_table, failures=failures)

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        task3 = self.get_task('Task3')
        task4 = self.get_task('Task4')

        self.assertEqual(task3.node_args, node_args)
        self.assertEqual(task4.node_args, node_args)

        # Now let's finish Task3 and Task4 to see that they correctly continue
        self.set_finished(task3)
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertIn('Task6', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 3)
        self.assertEqual(state_dict['waiting_edges'][0], 0)
        self.assertIn(2, state_dict['waiting_edges'])
        self.assertIn(3, state_dict['waiting_edges'])
        self.assertIn('Task3', state_dict['finished_nodes'])
        self.assertIn('Task4', state_dict['finished_nodes'])

        # Finish the flow!
        task5 = self.get_task('Task5')
        task6 = self.get_task('Task6')

        self.set_finished(task5)
        self.set_finished(task6)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        self.assertIsNone(retry)

    def test_single_failure_flow_fallback_to_flow(self):
        #
        # flow1:
        #
        #     Task1 X .... flow2
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['Task3'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {}, 'fallback': ['flow2']}}
        }
        self.init(edge_table, failures=failures)

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('flow2', self.instantiated_flows)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertEqual(system_state.node_args, node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

    def test_singe_failure_flow_fallback(self):
        #
        # flow1:
        #
        #     flow2 X ... Task2
        #       |
        #       |
        #     Task1
        #
        # Note:
        #   flow2 will fail
        #
        edge_table = {
            'flow1': [{'from': ['flow2'], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        failures = {
            'flow1': {'flow2': {'next:': {}, 'fallback': ['Task2']}}
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertNotIn('Task1', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

        # flow2 has failed
        flow2 = self.get_flow('flow2')
        flow_info = {'finished_nodes': {'Task1': ['id_task1']}, 'failed_nodes': {'Task2': ['id_task21']}}
        self.set_failed(flow2, FlowError(flow_info))

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertEqual(state_dict['waiting_edges'][0], 0)

    def test_two_failures_no_fallback(self):
        #
        # flow1:
        #
        #    Task1 X     Task2 X
        #       |           |
        #       |           |
        #    Task3        Task4
        #
        # Note:
        #   Task1 will fail, then Task2 will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
        }
        failures = {
            'flow1': {'Task1': {'next': {'Task2': {'next': {}, 'fallback': []}}, 'fallback': []},
                      'Task2': {'next': {'Task1': {'next': {}, 'fallback': []}}, 'fallback': []}
                     }
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        task1 = self.get_task('Task1')
        self.set_failed(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertIn('Task1', state_dict['failed_nodes'])

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertIn('Task1', state_dict['failed_nodes'])

        # Task2 has failed
        task2 = self.get_task('Task2')
        self.set_failed(task2)

        with self.assertRaises(FlowError):
            system_state = SystemState(id(self), 'flow1', state=state_dict,
                                       node_args=system_state.node_args)
            system_state.update()

    def test_two_failures_fallback_start(self):
        #
        # flow1:
        #
        #    Task1 X     Task2 X
        #       |           |
        #       |           |
        #    Task3        Task4
        #
        # Note:
        #   Task1 will fail, then Task2 will fail; fallback for Task1, Task2 is Task5
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
        }
        failures = {
            'flow1': {'Task1': {'next': {'Task2': {'next': {}, 'fallback': ['Task5']}}, 'fallback': []},
                      'Task2': {'next': {'Task1': {'next': {}, 'fallback': ['Task5']}}, 'fallback': []}
                      }
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertIn('Task1', state_dict['failed_nodes'])

        # Task2 has failed
        task2 = self.get_task('Task2')
        self.set_failed(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertNotIn('Task1', state_dict['failed_nodes'])
        self.assertNotIn('Task2', state_dict['failed_nodes'])

    def test_single_failure_finish_wait(self):
        #
        # flow1:
        #
        #    Task1       Task2 X
        #       |           |
        #       |           |
        #    Task3        Task4
        #
        # No fallback defined for Task2
        #
        # Note:
        #   Task2 will fail, no fallback defined, Dispatcher should wait for Task1 to finish and
        #   raise FlowError exception
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
        }
        failures = {
            'flow1': {'Task1': {'next': {'Task2': {'next': {}, 'fallback': []}}, 'fallback': []},
                      'Task2': {'next': {'Task1': {'next': {}, 'fallback': []}}, 'fallback': []}
                      }
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])

        # Task2 has failed
        task2 = self.get_task('Task2')
        self.set_failed(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertIn('Task2', state_dict['failed_nodes'])

        # No change so far, still wait
        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)
        self.assertNotIn('Task5', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 2)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertIn(1, state_dict['waiting_edges'])
        self.assertIn('Task2', state_dict['failed_nodes'])

        # Task1 has finished successfully
        task1 = self.get_task('Task1')
        self.set_finished(task1, 0)

        # Wait for Task3
        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertNotIn('Task4', self.instantiated_tasks)

        # Task3 has finished successfully
        task3 = self.get_task('Task3')
        self.set_finished(task3, 0)

        with self.assertRaises(FlowError):
            system_state = SystemState(id(self), 'flow1', state=state_dict,
                                       node_args=system_state.node_args)
            system_state.update()

    def test_propagate_finished_in_failure(self):
        #
        # flow1:
        #
        #    flow2 X      flow4 X    Task1_f1    Task2_f1 X   flow5
        #       |           |           |         |             |
        #        ................................................
        #                             |
        #                           TaskX
        #
        # flow2:
        #
        #     Task1_f2    Task2_f2 X   Task3_f2   flow3
        #
        # flow3:
        #
        #     Task1_f3    Task2_f3 X   Task3_f3 X
        #
        # flow4:
        #
        #     Task1_f4    Task2_f4 X   Task3_f4
        #
        # flow5:
        #     Task1_f5    Task2_f5 X
        #
        # Note:
        #  All flows fail, we inspect finished propagation in flow1 in case of failure. Tasks marked with X will fail.
        #
        edge_table = {
            'flow1': [{'from': ['flow2', 'flow5', 'flow4', 'Task1_f1', 'Task2_f1'], 'to': ['TaskX'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1_f1', 'Task2_f1', 'flow2', 'flow4', 'flow5'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['Task1_f2', 'Task2_f2', 'Task3_f2', 'flow3'], 'condition': self.cond_true}],
            'flow3': [{'from': [], 'to': ['Task1_f3', 'Task2_f3', 'Task3_f3'], 'condition': self.cond_true}],
            'flow4': [{'from': [], 'to': ['Task1_f4', 'Task2_f4', 'Task3_f4'], 'condition': self.cond_true}],
            'flow5': [{'from': [], 'to': ['Task1_f5', 'Task2_f5'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {
                'Task2_f1': {
                    'next': {
                        'flow2': {
                            'next': {
                                'flow4': {
                                    'next': {},
                                    'fallback': ['TaskX']
                                },
                                'fallback': []
                            },
                            'fallback': []
                        },
                        'fallback': []
                    },
                    'fallback': []
                },
                'fallback': []
            },
            'fallback': []
        }
        self.init(edge_table, failures=failures, propagate_finished={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertIn('flow4', self.instantiated_flows)
        self.assertIn('flow5', self.instantiated_flows)
        self.assertIn('Task1_f1', self.instantiated_tasks)
        self.assertIn('Task2_f1', self.instantiated_tasks)
        self.assertNotIn('flow3', self.instantiated_flows)

        # Task1_f1 have finished
        task1_f1 = self.get_task('Task1_f1')
        self.set_finished(task1_f1, 0)

        # Task2_f1 has failed
        task2_f1 = self.get_task('Task2_f1')
        self.set_failed(task2_f1)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertNotIn('flow3', self.instantiated_flows)

        # flow2

        system_state = SystemState(id(self), 'flow2')
        retry = system_state.update()
        state_dict_2 = system_state.to_dict()
        flow2 = self.get_flow('flow2')

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1_f2', self.instantiated_tasks)
        self.assertIn('Task2_f2', self.instantiated_tasks)
        self.assertIn('Task3_f2', self.instantiated_tasks)
        self.assertIn('flow3', self.instantiated_flows)

        # Task1_f2, Task3_f2 have finished
        task1_f2 = self.get_task('Task1_f2')
        self.set_finished(task1_f2, 1)

        task3_f2 = self.get_task('Task3_f2')
        self.set_finished(task3_f2, 2)

        # Task2_f2 has failed
        task2_f2 = self.get_task('Task2_f2')
        self.set_failed(task2_f2)

        system_state = SystemState(id(self), 'flow2', state=state_dict_2, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict_2 = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)

        # flow 3

        system_state = SystemState(id(self), 'flow3')
        retry = system_state.update()
        state_dict_3 = system_state.to_dict()
        flow3 = self.get_flow('flow3')

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1_f3', self.instantiated_tasks)
        self.assertIn('Task2_f3', self.instantiated_tasks)
        self.assertIn('Task3_f3', self.instantiated_tasks)

        # Task1_f3 has finished
        task1_f3 = self.get_task('Task1_f3')
        self.set_finished(task1_f3, 3)

        # Task2_f3, Task2_f3 have failed
        task2_f3 = self.get_task('Task2_f3')
        self.set_failed(task2_f3)

        task3_f3 = self.get_task('Task3_f3')
        self.set_failed(task3_f3)

        try:
            system_state = SystemState(id(self), 'flow3', state=state_dict_3, node_args=system_state.node_args)
            system_state.update()
        except FlowError as exc:
            self.set_failed(flow3, exc)
        else:
            self.assertFalse("Expected FlowError not raised")

        # flow4

        system_state = SystemState(id(self), 'flow4')
        retry = system_state.update()
        state_dict_4 = system_state.to_dict()
        flow4 = self.get_flow('flow4')

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1_f4', self.instantiated_tasks)
        self.assertIn('Task2_f4', self.instantiated_tasks)
        self.assertIn('Task3_f4', self.instantiated_tasks)

        # Task1_f2, Task3_f2 have finished
        task1_f4 = self.get_task('Task1_f4')
        self.set_finished(task1_f4, 1)

        task3_f4 = self.get_task('Task3_f4')
        self.set_finished(task3_f4, 4)

        # Task2_f4 has failed
        task2_f4 = self.get_task('Task2_f4')
        self.set_failed(task2_f4)

        try:
            system_state = SystemState(id(self), 'flow4', state=state_dict_4, node_args=system_state.node_args)
            system_state.update()
        except FlowError as exc:
            self.set_failed(flow4, exc)
        else:
            self.assertFalse("Expected FlowError not raised")

        # flow5

        system_state = SystemState(id(self), 'flow5')
        retry = system_state.update()
        state_dict_5 = system_state.to_dict()
        flow5 = self.get_flow('flow5')

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1_f5', self.instantiated_tasks)
        self.assertIn('Task2_f5', self.instantiated_tasks)

        # Task1_f5, Task2_f5 have finished
        task1_f5 = self.get_task('Task1_f5')
        self.set_finished(task1_f5, 1)
        task2_f5 = self.get_task('Task2_f5')
        self.set_finished(task2_f5, 2)

        system_state = SystemState(id(self), 'flow5', state=state_dict_5, node_args=system_state.node_args)
        retry = system_state.update()

        self.assertIsNone(retry)
        state_info = {
            'finished_nodes': system_state.to_dict()['finished_nodes'],
            'failed_nodes': system_state.to_dict()['failed_nodes']
        }
        self.set_finished(flow5, state_info)

        # Back to flow2, all child finished

        try:
            system_state = SystemState(id(self), 'flow2', state=state_dict_2, node_args=system_state.node_args)
            system_state.update()
        except FlowError as exc:
            self.set_failed(flow2, exc)
        else:
            self.assertFalse("Expected FlowError not raised")

        # finally inspect flow1 result and finished that is propagated at fallback to TaskX

        self.assertNotIn('TaskX', self.instantiated_tasks)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        system_state.update()

        self.assertIn('TaskX', self.instantiated_tasks)

        expected_finished = {
            'flow2': {
                'Task1_f2': [task1_f2.task_id],
                'Task3_f2': [task3_f2.task_id],
                'flow3': {
                    'Task1_f3': [task1_f3.task_id]
                }
            },
            'flow4': {
                'Task1_f4': [task1_f4.task_id],
                'Task3_f4': [task3_f4.task_id]
            },
        }

        expected_parent = {
            'Task2_f1': task2_f1.task_id,
            'flow2': {
                'Task2_f2': [task2_f2.task_id],
                'flow3': {
                    'Task2_f3': [task2_f3.task_id],
                    'Task3_f3': [task3_f3.task_id]
                }
            },
            'flow4': {
                'Task2_f4': [task2_f4.task_id]
            }
        }

        task_x = self.get_task('TaskX')

        self.assertEqual(expected_parent, task_x.parent)
        self.assertEqual(expected_finished, task_x.finished)



#!/usr/bin/env python3

from .selinonTask import SelinonTask
from .config import Config
from .dispatcher import Dispatcher
from .flowError import FlowError
from .storagePool import StoragePool
from .systemState import SystemState
from .version import selinon_version
from .fatalTaskError import FatalTaskError




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from .config import Config
from .lockPool import LockPool
from .trace import Trace


class StoragePool(object):
    """
    A pool that carries all database connections for workers
    """
    def __init__(self, id_mapping=None):
        self._id_mapping = id_mapping if id_mapping else {}

    @classmethod
    def get_storage_name_by_task_name(cls, task_name, graceful=False):
        """
        :param task_name: name of a task
        :param graceful: return None instead of raising an exception
        :return: storage name for task
        """
        try:
            return Config.task_mapping[task_name]
        except KeyError:
            if graceful:
                return None
            else:
                raise

    @classmethod
    def get_storage_by_task_name(cls, task_name):
        """
        :param task_name: task's name for which storage should be get
        :rtype: DataStorage
        """
        storage_name = cls.get_storage_name_by_task_name(task_name, graceful=True)
        if storage_name:
            return cls.get_connected_storage(storage_name)
        else:
            return None

    @classmethod
    def get_connected_storage(cls, storage_name):
        """
        Retrieve connected storage based by its name stated in configuration

        :param storage_name: name of storage
        :return: connected storage
        """
        # if this raises KeyError exception it means that the flow was not configured properly - should
        # be handled by Selinonlib
        storage = Config.storage_mapping[storage_name]

        if not storage.is_connected():
            with LockPool.get_lock(storage):
                if not storage.is_connected():
                    Trace.log(Trace.STORAGE_CONNECT, {'storage_name': storage_name})
                    storage.connect()

        return storage

    def get(self, task_name):
        """
        Retrieve data for task based on mapping for the current context

        :param task_name: task's name that we are retrieving data for
        :return: task's result for the current context
        """
        return self.retrieve(task_name, self._id_mapping[task_name])

    @classmethod
    def retrieve(cls, task_name, task_id):
        """
        Retrieve task's result from database which was configured to be used for desired task

        :param task_name: name of task for which result should be retrieved
        :param task_id: task ID to uniquely identify task results
        :return: task's result
        """
        storage = cls.get_storage_by_task_name(task_name)
        Trace.log(Trace.STORAGE_RETRIEVE, {'task_name': task_name,
                                           'storage_name': Config.task_mapping[task_name]})
        return storage.retrieve(task_name, task_id)

    @classmethod
    def set(cls, node_args, flow_name, task_name, task_id, result):
        """
        Store result for task

        :param node_args: arguments that were passed to the node
        :param flow_name: flow in which task was run
        :param task_name: task that computed result
        :param task_id: task id that computed result
        :param result: result that should be stored
        :return: result ID - a unique ID which can be used to reference task results
        """
        storage = cls.get_storage_by_task_name(task_name)

        record_id = storage.store(node_args, flow_name, task_name, task_id, result)
        Trace.log(Trace.STORAGE_STORE, {'flow_name': flow_name,
                                        'node_args': node_args,
                                        'task_name': task_name,
                                        'task_id': task_id,
                                        'storage_name': Config.task_mapping[task_name],
                                        'result': result,
                                        'record_id': record_id})
        return record_id



#!/usr/bin/env python3

from sqlalchemy import create_engine, Column, Integer, Sequence, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import create_database, database_exists
from sqlalchemy.orm import sessionmaker
from .dataStorage import DataStorage

Base = declarative_base()


class Result(Base):
    __tablename__ = 'result'

    id = Column(Integer, Sequence('result_id'), primary_key=True)
    flow_name = Column(String(128))
    task_name = Column(String(128))
    task_id = Column(String(255), unique=True)
    # We are using JSONB for postgres, if you want to use other database, change column type
    result = Column(JSONB)
    node_args = Column(JSONB)

    def __init__(self, node_args, flow_name, task_name, task_id, result):
        self.flow_name = flow_name
        self.task_name = task_name
        self.task_id = task_id
        self.result = result
        self.node_args = node_args


class SqlStorage(DataStorage):
    def __init__(self, connection_string, encoding='utf-8', echo=False):
        super(SqlStorage, self).__init__()

        self.engine = create_engine(connection_string, encoding=encoding, echo=echo)
        self.session = None

    def is_connected(self):
        return self.session is not None

    def connect(self):
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        self.session = sessionmaker(bind=self.engine)()
        Base.metadata.create_all(self.engine)

    def disconnect(self):
        if self.is_connected():
            self.session.close()
            self.session = None

    def retrieve(self, task_name, task_id):
        assert(self.is_connected())

        record = self.session.query(Result).filter_by(task_id=task_id).one()

        assert(record.task_name == task_name)
        return record.result

    def store(self, node_args, flow_name, task_name, task_id, result):
        assert(self.is_connected())

        record = Result(node_args, flow_name, task_name, task_id, result)
        try:
            self.session.add(record)
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

        return record.id




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from threading import Lock


class LockPool(object):
    """
    Global lock pool for shared locks
    """
    _locks = {}
    _global_lock = Lock()

    @classmethod
    def get_lock(cls, lock_id):
        """
        Get lock for resource, exclusively

        :param lock_id: lock_id for uniquely identify lock
        :return: lock, can be acquired if already taken, if new, always released
        """
        # TODO: we could make it more optimal by storing only locks that are used and release them once unused
        with cls._global_lock:
            if not cls._locks.get(lock_id):
                ret = Lock()
                cls._locks[lock_id] = ret
            else:
                ret = cls._locks[lock_id]

        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import traceback
from celery import Task
from .systemState import SystemState
from .flowError import FlowError
from .trace import Trace
from .config import Config


class Dispatcher(Task):
    """
    Selinon Dispatcher worker implementation
    """
    # Celery configuration
    ignore_result = False
    acks_late = True
    track_started = True
    max_retries = None
    name = "Dispatcher"

    def run(self, flow_name, node_args=None, parent=None, finished=None, retry=None, state=None):
        """
        Dispatcher entry-point - run each time a dispatcher is scheduled

        :param flow_name: name of the flow
        :param parent: flow parent nodes
        :param node_args: arguments for workers
        :param finished: finished tasks in case of fallback is run, otherwise None
        :param retry: last retry countdown
        :param state: the current system state
        :raises: FlowError
        """

        Trace.log(Trace.DISPATCHER_WAKEUP, {'flow_name': flow_name,
                                            'dispatcher_id': self.request.id,
                                            'node_args': node_args,
                                            'finished': finished,
                                            'retry': retry,
                                            'state': state})
        try:
            system_state = SystemState(self.request.id, flow_name, node_args, retry, state, parent, finished)
            retry = system_state.update()
        except FlowError as flow_error:
            Trace.log(Trace.FLOW_FAILURE, {'flow_name': flow_name,
                                           'dispatcher_id': self.request.id,
                                           'state': state,
                                           'node_args': node_args,
                                           'parent': parent,
                                           'finished': finished,
                                           'retry': retry,
                                           'what': str(flow_error)})
            # force max_retries to 0 so we are not scheduled and marked as FAILED
            raise self.retry(max_retries=0, exc=flow_error)
        except Exception as exc:
            Trace.log(Trace.DISPATCHER_FAILURE, {'flow_name': flow_name,
                                                 'dispatcher_id': self.request.id,
                                                 'what': traceback.format_exc()})
            raise

        state_dict = system_state.to_dict()
        node_args = system_state.node_args

        if retry:
            kwargs = {
                'flow_name': flow_name,
                'node_args': node_args,
                'parent': parent,
                'retry': retry,
                'state': state_dict
            }
            Trace.log(Trace.DISPATCHER_RETRY, {'flow_name': flow_name,
                                               'dispatcher_id': self.request.id,
                                               'retry': retry,
                                               'state_dict': state_dict,
                                               'node_args': node_args
                                               })
            raise self.retry(args=[], kwargs=kwargs, countdown=retry, queue=Config.dispatcher_queue)
        else:
            Trace.log(Trace.FLOW_END, {'flow_name': flow_name,
                                       'dispatcher_id': self.request.id,
                                       'finished_nodes': state_dict['finished_nodes']})
            return {
                'finished_nodes': state_dict['finished_nodes'],
                # this is always {} since we have finished, but leave it here because of failure tracking
                'failed_nodes': state_dict['failed_nodes']
            }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import traceback
import jsonschema
from .storagePool import StoragePool
from .fatalTaskError import FatalTaskError
from celery import Task
from .trace import Trace
from .config import Config


class SelinonTaskEnvelope(Task):
    """
    A Celery task that is responsible for running user defined tasks from flow
    """
    # Celery configuration
    ignore_result = False
    acks_late = True
    track_started = True
    max_retries = None
    name = "SelinonTaskEnvelope"

    @classmethod
    def validate_result(cls, task_name, result):
        """
        :param task_name:
        :param result:
        :return:
        """
        schema_path = Config.output_schemas.get(task_name)
        if schema_path:
            with open(schema_path, "r") as f:
                schema = f.read()

            jsonschema.validate(result, schema)

    def run(self, task_name, flow_name, parent, node_args, finished=None, retried_count=None):
        """
        :param task_name:
        :param flow_name:
        :param parent:
        :param node_args:
        :param finished:
        :param retried_count:
        :return:
        """
        # we are passing args as one argument explicitly for now not to have troubles with *args and **kwargs mapping
        # since we depend on previous task and the result can be anything
        Trace.log(Trace.TASK_START, {'flow_name': flow_name,
                                     'task_name': task_name,
                                     'task_id': self.request.id,
                                     'parent': parent,
                                     'finished': finished,
                                     'args': node_args})
        try:
            task = Config.get_task_instance(task_name=task_name, flow_name=flow_name, parent=parent, finished=finished)
            result = task.run(node_args)
            self.validate_result(task_name, result)

            storage = StoragePool.get_storage_by_task_name(task_name)
            if storage and not Config.storage_readonly[task_name]:
                storage.store(node_args, flow_name, task_name, self.request.id, result)
            elif result is not None:
                Trace.log(Trace.TASK_DISCARD_RESULT, {'flow_name': flow_name,
                                                      'task_name': task_name,
                                                      'task_id': self.request.id,
                                                      'parent': parent,
                                                      'args': node_args,
                                                      'finished': finished,
                                                      'result': result})
        except Exception as exc:
            max_retry = Config.max_retry.get(task_name, 0)
            retried_count = 0 if retried_count is None else retried_count

            if max_retry > retried_count and not isinstance(exc, FatalTaskError):
                retried_count += 1
                retry_countdown = Config.retry_countdown.get(task_name, 0)

                kwargs = {
                    'task_name': task_name,
                    'flow_name': flow_name,
                    'parent': parent,
                    'node_args': node_args,
                    'retried_count': retried_count
                }

                Trace.log(Trace.TASK_RETRY, {'flow_name': flow_name,
                                             'task_name': task_name,
                                             'task_id': self.request.id,
                                             'parent': parent,
                                             'finished': finished,
                                             'args': node_args,
                                             'what': traceback.format_exc(),
                                             'retry_countdown': retry_countdown,
                                             'retried_count': retried_count,
                                             'max_retry': max_retry})
                raise self.retry(kwargs=kwargs, countdown=retry_countdown, queue=Config.task_queues[task_name])
            else:
                Trace.log(Trace.TASK_FAILURE, {'flow_name': flow_name,
                                               'task_name': task_name,
                                               'task_id': self.request.id,
                                               'parent': parent,
                                               'finished': finished,
                                               'args': node_args,
                                               'what': traceback.format_exc(),
                                               'retried_count': retried_count})
                raise self.retry(max_retries=0, exc=exc)

        Trace.log(Trace.TASK_END, {'flow_name': flow_name,
                                   'task_name': task_name,
                                   'task_id': self.request.id,
                                   'parent': parent,
                                   'finished': finished,
                                   'args': node_args,
                                   'storage': StoragePool.get_storage_name_by_task_name(task_name, graceful=True)})




#!/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import itertools
import json
from functools import reduce
from celery.result import AsyncResult
from .flowError import FlowError
from .storagePool import StoragePool
from .trace import Trace
from .config import Config
from .selinonTaskEnvelope import SelinonTaskEnvelope


class SystemState(object):
    """
    Main system actions done by Selinon
    """

    @property
    def node_args(self):
        """
        :return: arguments for a node
        """
        return self._node_args

    @staticmethod
    def _instantiate_active_nodes(arr):
        """
        :return: convert node references from argument to AsyncResult
        """
        return [{'name': node['name'], 'id': node['id'], 'result': AsyncResult(id=node['id'])} for node in arr]

    @staticmethod
    def _deinstantiate_active_nodes(arr):
        """
        :return: node references for Dispatcher arguments
        """
        return [{'name': x['name'], 'id': x['id']} for x in arr]

    @staticmethod
    def _idxs2items(edge_node_table, arr):
        # we keep only the index to edges_table in order to avoid serialization and optimize number representation
        ret = []
        for i in arr:
            ret.append(edge_node_table[i])
        return ret

    def __init__(self, dispatcher_id, flow_name, node_args = None, retry = None, state = None, parent=None,
                 finished=None):
        state_dict = {} if state is None else state

        self._dispatcher_id = dispatcher_id
        self._flow_name = flow_name
        self._node_args = node_args
        self._parent = parent if parent else {}
        self._finished = finished
        self._active_nodes = self._instantiate_active_nodes(state_dict.get('active_nodes', []))
        self._finished_nodes = state_dict.get('finished_nodes', {})
        self._failed_nodes = state_dict.get('failed_nodes', {})
        self._waiting_edges_idx = state_dict.get('waiting_edges', [])
        # Instantiate lazily later if we will know that there is something to process
        self._waiting_edges = []
        self._retry = retry

    def to_dict(self):
        """
        :return: converted system state to dict
        """
        return {'active_nodes': self._deinstantiate_active_nodes(self._active_nodes),
                'finished_nodes': self._finished_nodes,
                'failed_nodes': self._failed_nodes,
                'waiting_edges': self._waiting_edges_idx
                }

    def _get_successful_and_failed(self):
        """
        :return: all successful and failed nodes in system from active nodes
        """
        ret_successful = []
        ret_failed = []

        new_active_nodes = []
        for node in self._active_nodes:
            if node['result'].successful():
                Trace.log(Trace.NODE_SUCCESSFUL, {'flow_name': self._flow_name,
                                                  'dispatcher_id': self._dispatcher_id,
                                                  'node_name': node['name'],
                                                  'node_id': node['id']})
                ret_successful.append(node)
            elif node['result'].failed():
                Trace.log(Trace.NODE_FAILURE, {'flow_name': self._flow_name,
                                               'dispatcher_id': self._dispatcher_id,
                                               'node_name': node['name'],
                                               'node_id': node['id'],
                                               'what': node['result'].result})
                # We keep track of failed nodes to handle failures once all nodes finish
                if node['name'] not in self._failed_nodes:
                    self._failed_nodes[node['name']] = []
                self._failed_nodes[node['name']].append(node['id'])
                ret_failed.append(node['id'])
            else:
                new_active_nodes.append(node)

        self._active_nodes = new_active_nodes

        return ret_successful, ret_failed

    def _start_node(self, node_name, parent, node_args, finished=None):
        """
        Start a node in the system

        :param node_name: name of a node to be started
        :param parent: parent nodes of the starting node
        :param node_args: arguments for the starting node
        :param finished: finished nodes for starting node in case of subflow and propagate_finished flag
        """
        from .dispatcher import Dispatcher
        if Config.is_flow(node_name):
            if Config.propagate_node_args.get(self._flow_name):
                if Config.propagate_node_args.get(self._flow_name) is True or \
                    (isinstance(Config.propagate_node_args.get(self._flow_name), list) and
                     node_name in Config.propagate_node_args.get(self._flow_name)):
                    node_args = node_args
                else:
                    node_args = None
            else:
                node_args = None

            if Config.propagate_parent.get(self._flow_name):
                if Config.propagate_parent.get(self._flow_name) is True or \
                    (isinstance(Config.propagate_parent.get(self._flow_name), list) and
                     node_name in Config.propagate_parent.get(self._flow_name)):
                    parent = parent
                else:
                    parent = None
            else:
                parent = None

            kwargs = {
                'flow_name': node_name,
                'node_args': node_args,
                'parent': parent,
                'finished': finished
            }

            async_result = Dispatcher().apply_async(kwargs=kwargs, queue=Config.dispatcher_queue)

            # reuse kwargs for trace log entry
            kwargs['flow_name'] = self._flow_name
            kwargs['child_flow_name'] = node_name
            kwargs['dispatcher_id'] = self._dispatcher_id
            kwargs['child_dispatcher_id'] = async_result.task_id,
            kwargs['queue'] = Config.dispatcher_queue
            Trace.log(Trace.SUBFLOW_SCHEDULE, kwargs)
        else:
            kwargs = {
                'task_name': node_name,
                'flow_name': self._flow_name,
                'parent': parent,
                'node_args': node_args,
                'finished': finished
            }

            async_result = SelinonTaskEnvelope().apply_async(kwargs=kwargs, queue=Config.task_queues[node_name])

            # reuse kwargs for trace log entry
            kwargs['dispatcher_id'] = self._dispatcher_id
            kwargs['task_id'] = async_result.task_id
            Trace.log(Trace.TASK_SCHEDULE, kwargs)

        record = {'name': node_name, 'id': async_result.task_id, 'result': async_result}

        if node_name not in Config.nowait_nodes.get(self._flow_name, {}):
            self._active_nodes.append(record)

        return record

    def _run_fallback(self):
        """
        Run fallback in the system

        :return: fallbacks that were run
        """
        # we sort it first to make evaluation dependent on alphabetical order
        # TODO: use binary search when inserting to optimize from O(N*log(N)) to O(log(N))
        ret = []
        failed_nodes = sorted(self._failed_nodes.items())

        for i in range(len(failed_nodes), 0, -1):
            for combination in itertools.combinations(failed_nodes, i):
                try:
                    failure_nodes = Config.failures[self._flow_name]
                    failure_node = reduce(lambda n, c: n['next'][c[0]], combination[1:], failure_nodes[combination[0][0]])
                except KeyError:
                    # such failure not found in the tree of permutations - this means that this flow will always fail,
                    # but run defined fallbacks for defined failures first
                    continue

                if isinstance(failure_node['fallback'], list) and len(failure_node['fallback']) > 0:
                    parent = {}
                    finished = {}
                    traced_nodes_arr = []

                    for node in combination:
                        traced_nodes_arr.append((node[0], self._failed_nodes[node[0]],))

                        if Config.is_flow(node[0]) and Config.propagate_finished:
                            # we have to add all subsequent finished nodes to list to track info about finished
                            task_id = self._failed_nodes[node[0]][0]
                            # Celery stores exception that was raised in case of failure in result property
                            # We keep there information about finished and failed nodes
                            raw_result = str(AsyncResult(task_id).result)
                            flow_info = json.loads(raw_result)
                            parent[node[0]] = {}
                            finished[node[0]] = {}
                            self._extend_parent_finished(parent[node[0]], finished[node[0]], flow_info)
                            self._failed_nodes[node[0]].pop(0)
                        else:
                            parent[node[0]] = self._failed_nodes[node[0]].pop(0)

                        if len(self._failed_nodes[node[0]]) == 0:
                            del self._failed_nodes[node[0]]

                    Trace.log(Trace.FALLBACK_START, {'flow_name': self._flow_name,
                                                     'dispatcher_id': self._dispatcher_id,
                                                     'nodes': traced_nodes_arr,
                                                     'fallback': failure_node['fallback']})

                    for node in failure_node['fallback']:
                        record = self._start_node(node, parent=parent, node_args=self._node_args,
                                                  finished=finished)
                        ret.append(record)

                    # wait for fallback to finish in order to avoid time dependent flow evaluation
                    return ret
                elif failure_node['fallback'] is True:
                    traced_nodes_arr = []
                    for node in combination:
                        traced_nodes_arr.append((node[0], self._failed_nodes[node[0]],))
                        self._failed_nodes[node[0]].pop(0)
                        if len(self._failed_nodes[node[0]]) == 0:
                            del self._failed_nodes[node[0]]

                    Trace.log(Trace.FALLBACK_START, {'flow_name': self._flow_name,
                                                     'dispatcher_id': self._dispatcher_id,
                                                     'nodes': traced_nodes_arr,
                                                     'fallback': True})
                    # continue with fallback in other combinations, nothing started

        return ret

    def _update_waiting_edges(self, node_name):
        """
        Update waiting edges (edges that wait for a node finish) based on node name

        :param node_name: node that will trigger an edge
        """
        for idx, edge in enumerate(Config.edge_table[self._flow_name]):
            if node_name in edge['from'] and idx not in self._waiting_edges_idx:
                self._waiting_edges.append(edge)
                self._waiting_edges_idx.append(idx)

    def _extend_parent_finished(self, parent_dict, finished_dict, flow_info):
        """"
        Compute finished and parent for starting node in case of propagate_finished flag

        :param parent_dict: parent dict that should be extended
        :param finished_dict: finished dict that should be extended
        :param flow_info: flow info propagated from a flow
        """
        for node_name, node_val in flow_info['finished_nodes'].items():
            if Config.is_flow(node_name):
                finished_dict[node_name] = {}
                if node_name not in parent_dict:
                    parent_dict[node_name] = {}
                for node_id in node_val:
                    raw_result = str(AsyncResult(node_id).result)
                    child_flow_info = json.loads(raw_result)
                    self._extend_parent_finished(parent_dict[node_name], finished_dict[node_name], child_flow_info)
            else:
                finished_dict[node_name] = node_val

        for node_name, node_val in flow_info['failed_nodes'].items():
            if Config.is_flow(node_name):
                parent_dict[node_name] = {}
                if node_name not in finished_dict:
                    finished_dict[node_name] = {}
                for node_id in node_val:
                    result = AsyncResult(node_id).result

                    if not isinstance(result, FlowError):
                        # If we had some exception that was caused not by error in flow (e.g. bug in dispatcher, failed
                        # to connect to db, ...) we propagate it
                        raise result

                    child_flow_info = json.loads(str(result))
                    self._extend_parent_finished(parent_dict[node_name], finished_dict[node_name], child_flow_info)
            else:
                parent_dict[node_name] = node_val

    def _extend_parent_from_flow(self, parent_dict, flow_id):
        """
        Compute parent in a flow in case of propagate_parent flag

        :param parent_dict: parent dict that should be extended
        :param flow_id: flow id that should be inspected and used to extend parent_dict
        """
        async_result = AsyncResult(flow_id)

        for node_name, node_ids in async_result.result['finished_nodes'].items():
            if Config.is_flow(node_name):
                parent_dict[node_name] = {}
                for node_id in node_ids:
                    self._extend_parent_from_flow(parent_dict[node_name], node_id)
            else:
                if node_name not in parent_dict:
                    parent_dict[node_name] = []
                for node_id in node_ids:
                    parent_dict[node_name].append(node_id)

    def _start_new_from_finished(self, new_finished):
        """
        Start new based on finished nodes

        :param new_finished: finished nodes based on which we should start new nodes
        :return: newly started nodes
        """
        ret = []

        if not self._node_args and len(new_finished) == 1 and len(self._active_nodes) == 0 and \
                len(self._finished_nodes) == 0:
            # propagate arguments from newly finished node only if:
            #  1. we did not define node arguments in original Dispatcher() call
            #  2. the flow started only with a one single node that just finished
            #  3. node was not a subflow
            # TODO: make this configurable from YAML
            if not Config.is_flow(new_finished[0]['name']):
                self._node_args = new_finished[0]['result'].result

        for node in new_finished:
            # We could optimize this by pre-computing affected edges in pre-generated config file for each
            # node and computing intersection with waiting edges, but let's stick with this solution for now
            edges = [edge for edge in self._waiting_edges if node['name'] in edge['from']]

            for edge in edges:
                from_nodes = dict.fromkeys(edge['from'], [])

                for from_name in from_nodes:
                    if from_name == node['name']:
                        from_nodes[from_name] = [{'name': node['name'], 'id': node['id']}]
                    else:
                        from_nodes[from_name] = [{'name': from_name, 'id': n}
                                                 for n in self._finished_nodes.get(from_name, [])]

                # if there are multiple nodes of a same type waiting on an edge:
                #
                #   A    B
                #    \  /
                #     \/
                #     C
                #
                # A: <id1>, <id2>
                # B: <id3>, <id4>
                #
                # Compute all combinations: (<id1>, <id3>), (<id1>, <id4>), (<id2>, <id3>), (<id2>, <id4>)
                #
                for start_nodes in itertools.product(*from_nodes.values()):
                    parent = {}
                    storage_id_mapping = {}

                    for start_node in start_nodes:
                        if Config.is_flow(start_node['name']):
                            propagate_finished = Config.propagate_finished.get(self._flow_name, False)
                            if propagate_finished is True or (isinstance(propagate_finished, list) and
                                                              start_node['name'] in propagate_finished):
                                parent[start_node['name']] = {}
                                self._extend_parent_from_flow(parent[start_node['name']], start_node['id'])
                        else:
                            parent[start_node['name']] = start_node['id']
                            storage_id_mapping[start_node['name']] = start_node['id']

                    # We could also examine results of subflow, there could be passed a list of subflows with
                    # finished_nodes to 'condition' in order to do inspection
                    storage_pool = StoragePool(storage_id_mapping)
                    if edge['condition'](storage_pool, self._node_args):
                        for node_name in edge['to']:
                            record = self._start_node(node_name, parent=parent, node_args=self._node_args,
                                                      finished=self._finished)
                            ret.append(record)

            node_name = node['name']
            if not self._finished_nodes.get(node_name):
                self._finished_nodes[node_name] = []
            self._finished_nodes[node_name].append(node['id'])

        return ret

    def _start_and_update_retry(self):
        """
        Start the flow and update retry

        :return: new/next retry
        """
        new_started_nodes = []

        Trace.log(Trace.FLOW_START, {'flow_name': self._flow_name,
                                     'dispatcher_id': self._dispatcher_id,
                                     'args': self._node_args})
        start_edges = [edge for edge in Config.edge_table[self._flow_name] if len(edge['from']) == 0]
        if len(start_edges) == 0:
            # This should not occur since selinonlib raises exception if such occurs, but just to be sure...
            raise ValueError("No starting node found for flow '%s'!" % self._flow_name)

        for start_edge in start_edges:
            storage_pool = StoragePool()
            if start_edge['condition'](storage_pool, self._node_args):
                for node_name in start_edge['to']:
                    node = self._start_node(node_name, node_args=self._node_args, parent=self._parent,
                                            finished=self._finished)
                    if node_name not in Config.nowait_nodes.get(self._flow_name, {}):
                        self._update_waiting_edges(node_name)
                        new_started_nodes.append(node)

        self._retry = Config.strategy_function(previous_retry=None,
                                               active_nodes=self._active_nodes,
                                               failed_nodes=self._failed_nodes,
                                               new_started_nodes=new_started_nodes,
                                               new_fallback_nodes=[])

    def update(self):
        """
        Check the current state in the system and start new nodes if possible

        :return: retry count - can be None (do not retry dispatcher) or time in seconds to retry
        """

        if not self._active_nodes and not self._finished_nodes and not self._waiting_edges and not self._failed_nodes:
            # we are starting up
            self._start_and_update_retry()
        else:
            new_finished, new_failed = self._get_successful_and_failed()

            if new_finished or new_failed:
                # Instantiate results lazily
                self._waiting_edges = self._idxs2items(Config.edge_table[self._flow_name], self._waiting_edges_idx)

                for node in new_finished:
                    self._update_waiting_edges(node['name'])

                started = self._start_new_from_finished(new_finished)

                fallback_started = []
                # We wait until all active nodes finish if there are some failed nodes try to recover from failure,
                # otherwise mark flow as failed
                if len(self._active_nodes) == 0 and self._failed_nodes:
                    fallback_started = self._run_fallback()
                    if len(fallback_started) == 0 and len(self._failed_nodes) > 0:
                        # We will propagate state information so we can correctly propagate parent and failed nodes in
                        # parent flow if there is any
                        # We use JSON in order to use result backend with JSON configured so objects are not pickled
                        state_dict = self.to_dict()
                        state_info = {
                            'finished_nodes': state_dict['finished_nodes'],
                            'failed_nodes': state_dict['failed_nodes']
                        }
                        raise FlowError(json.dumps(state_info))

                self._retry = Config.strategy_function(previous_retry=self._retry,
                                                       active_nodes=self._active_nodes,
                                                       failed_nodes=self._failed_nodes,
                                                       new_started_nodes=started,
                                                       new_fallback_nodes=fallback_started)
            else:
                self._retry = Config.strategy_function(previous_retry=self._retry,
                                                       active_nodes=self._active_nodes,
                                                       failed_nodes=self._failed_nodes,
                                                       new_started_nodes=[],
                                                       new_fallback_nodes=[])
        return self._retry



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from selinonTestCase import SelinonTestCase

from selinon import SystemState


class TestNodeFailures(SelinonTestCase):
    def test_nowait_task(self):
        #
        # flow1:
        #
        #     Task1
        #
        # Note:
        #    Task1 is marked as nowait node
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        nowait_nodes = {'flow1': ['Task1']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task1', state_dict.get('active_nodes'))
        self.assertEqual(len(state_dict.get('waiting_edges')), 0)

    def test_nowait_flow(self):
        #
        # flow1:
        #
        #     flow2
        #
        # Note:
        #    flow2 is marked as nowait node
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        nowait_nodes = {'flow1': ['flow2']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('flow2', self.instantiated_flows)
        self.assertNotIn('flow2', state_dict.get('active_nodes'))
        self.assertEqual(len(state_dict.get('waiting_edges')), 0)

    def test_nowait_in_flow(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task3 finishes before Task2 and Task3 is marked as nowait
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        nowait_nodes = {'flow1': ['Task3']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIsNone(system_state.node_args)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertNotIn('Task2', self.instantiated_tasks)
        self.assertNotIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 0)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        self.assertIsNotNone(retry)
        self.assertIn('Task1', self.instantiated_tasks)
        self.assertIn('Task2', self.instantiated_tasks)
        self.assertIn('Task3', self.instantiated_tasks)
        self.assertEqual(len(state_dict.get('waiting_edges')), 1)
        self.assertIn(0, state_dict['waiting_edges'])
        self.assertEqual(len(state_dict.get('finished_nodes')), 1)
        self.assertEqual(len(state_dict.get('active_nodes')), 1)
        self.assertNotIn('Task3', state_dict.get('active_nodes'))
        self.assertIn('Task2', state_dict['active_nodes'][0]['name'])



#!/usr/bin/env python3

import redis
import json
from .dataStorage import DataStorage


class RedisStorage(DataStorage):
    def __init__(self, host="localhost", port=6379, db=0, password=None, socket_timeout=None, connection_pool=None,
                 charset='utf-8', errors='strict', unix_socket_path=None):
        super(RedisStorage, self).__init__()
        self.conn = None
        self.host = host
        self.port = port
        self.db = db
        self.password = password
        self.socket_timeout = socket_timeout
        self.connection_pool = connection_pool
        self.charset = charset
        self.errors = errors
        self.unix_socket_path = unix_socket_path

    def is_connected(self):
        return self.conn is not None

    def connect(self):
        self.conn = redis.Redis(host=self.host, port=self.port, db=self.db, password=self.password,
                                socket_timeout=self.socket_timeout, connection_pool=self.connection_pool,
                                charset=self.charset, errors=self.errors)

    def disconnect(self):
        if self.is_connected():
            self.conn.connection_pool.disconnect()
            self.conn = None

    def retrieve(self, task_name, task_id):
        assert(self.is_connected())

        ret = self.conn.get(task_id)

        if ret is None:
            raise FileNotFoundError("Record not found in database")

        record = json.loads(ret.decode(self.charset))

        assert(record.get('task_name') == task_name)
        return record.get('result')

    def store(self, node_args, flow_name, task_name, task_id, result):
        assert(self.is_connected())

        record = {
            'node_args': node_args,
            'flow_name': flow_name,
            'task_name': task_name,
            'task_id': task_id,
            'result': result
        }

        self.conn.set(task_id, json.dumps(record))
        return task_id




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import abc
from .storagePool import StoragePool


class SelinonTask(metaclass=abc.ABCMeta):
    """
    Base class for user-defined tasks
    """
    def __init__(self, flow_name, task_name, parent, finished):
        """
        :param flow_name: name of flow under which this tasks runs on
        :param task_name: name of task, note it can be aliased since we can have different task name and class name
        :param parent: direct task's predecessors stated in flow dependency
        :param finished: tasks that were finished in parent subflow in case task is fallback from fallback
        """
        self.flow_name = flow_name
        self.task_name = task_name
        self.parent = parent
        self.finished = finished

    @property
    def storage(self):
        """
        :return: tasks's configured storage as stated in YAML config
        """
        return StoragePool.get_storage_by_task_name(self.task_name)

    def parent_task_result(self, parent_name):
        """
        :param parent_name: name of parent task to retrieve result from
        :return: result of parent task
        """
        return StoragePool.retrieve(parent_name, self.parent[parent_name])

    def finished_task_result(self, finished_name):
        """
        :param finished_name: name of finished node in parent subflow
        :return: result of finished task in parent subflow
        """
        # TODO: we should should distinguish sub-subflows here
        return StoragePool.retrieve(finished_name, self.finished[finished_name])

    def parent_flow_result(self, flow_name, task_name, index):
        """
        Get parent subflow results; note that parent flows can return multiple results from task of same type
        because of loops in flows

        :param flow_name: name of parent flow
        :param task_name: name of task in parent flow
        :param index: index of result if more than one subflow was run
        :return: result of task in parent subflow
        """
        # TODO: we should should distinguish sub-subflows here
        return StoragePool.retrieve(task_name, self.parent[flow_name][task_name][index])

    @abc.abstractmethod
    def run(self, node_args):
        """
        Task's entrypoint - user defined computation

        :param node_args: arguments passed to flow/node
        :return: tasks's result that will be stored in database as configured
        """
        pass




#!/usr/bin/python3

from setuptools import setup


def get_requirements():
    with open('requirements.txt') as fd:
        return fd.read().splitlines()


setup(
    name='selinon',
    version='0.1.0rc3',
    packages=['selinon', 'selinon.storage'],
    install_requires=get_requirements(),
    author='Fridolin Pokorny',
    author_email='fpokorny@redhat.com',
    maintainer='Fridolin Pokorny',
    maintainer_email='fpokorny@redhat.com',
    description='task flow management for Celery',
    url='https://github.com/fridex/selinon',
    license='GPL',
    keywords='celery selinonlib yaml condition flow',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import os
import runpy
import tempfile
from .trace import Trace
from selinonlib import System


class Config(object):
    """
    All user configurations generated from YAML file
    """
    celery_app = None

    get_task_instance = None
    is_flow = None
    edge_table = None
    failures = None
    nowait_nodes = None
    max_retry = None
    retry_countdown = None
    storage_readonly = None
    propagate_node_args = None
    propagate_finished = None
    propagate_parent = None
    propagate_compound_finished = None
    propagate_compound_parent = None
    output_schemas = None

    storage_mapping = None
    task_mapping = None
    dispatcher_queue = None
    task_queues = None
    strategy_function = None


    @classmethod
    def _set_config(cls, config_module):
        """
        Set configuration from Python's module
        """
        cls.get_task_instance = config_module['get_task_instance']
        cls.is_flow = config_module['is_flow']
        cls.edge_table = config_module['edge_table']
        cls.failures = config_module['failures']
        cls.nowait_nodes = config_module['nowait_nodes']

        # propagate_* entries
        cls.propagate_finished = config_module['propagate_finished']
        cls.propagate_node_args = config_module['propagate_node_args']
        cls.propagate_parent = config_module['propagate_parent']
        cls.propagate_compound_finished = config_module['propagate_compound_finished']
        cls.propagate_compound_parent = config_module['propagate_compound_parent']

        # task configuration
        cls.output_schemas = config_module['output_schemas']
        cls.storage_mapping = config_module['storage2instance_mapping']
        cls.task_mapping = config_module['task2storage_mapping']
        cls.max_retry = config_module['max_retry']
        cls.retry_countdown = config_module['retry_countdown']
        cls.storage_readonly = config_module['storage_readonly']

        # queues
        cls.dispatcher_queue = config_module['dispatcher_queue']
        cls.task_queues = config_module['task_queues']

        # Dispatcher scheduling strategy
        cls.strategy_function = config_module['strategy_function']

        # call config init with Config class to set up other configuration specific values
        config_module['init'](cls)

    @classmethod
    def set_config_py(cls, config_code):
        """
        Set dispatcher configuration by Python config file

        :param config_code: configuration source code
        """
        config_module = runpy.run_path(config_code)
        cls._set_config(config_module)

    @classmethod
    def set_config_yaml(cls, nodes_definition_file, flow_definition_files, config_py=None):
        """
        Set dispatcher configuration by path to YAML configuration files

        :param nodes_definition_file: definition of system nodes - YAML configuration
        :param flow_definition_files: list of flow definition files
        :param config_py: a file that should be used for storing generated config.py
        """
        system = System.from_files(nodes_definition_file, flow_definition_files)

        if not config_py:
            tmp_f = tempfile.NamedTemporaryFile(mode="w", delete=False)
            system.dump2stream(tmp_f)
            tmp_f.close()
            cls.set_config_py(tmp_f.name)
            os.unlink(tmp_f.name)
        else:
            with open(config_py, "w") as f:
                system.dump2stream(f)
            cls.set_config_py(config_py)

    @classmethod
    def trace_by_func(cls, trace_func):
        """
        Set tracing function for Dispatcher

        :param trace_func: a function that should be used to trace dispatcher actions
        """
        Trace.trace_by_func(trace_func)

    @classmethod
    def trace_by_logging(cls):
        """
        Use Python's logging for tracing
        """
        Trace.trace_by_logging()

    @classmethod
    def set_celery_app(cls, celery_app):
        """
        Set celery application that should be used

        :param celery_app: celery app instance
        """
        # Avoid circular imports
        from .dispatcher import Dispatcher
        from .selinonTaskEnvelope import SelinonTaskEnvelope

        cls.celery_app = celery_app
        celery_app.tasks.register(Dispatcher())
        celery_app.tasks.register(SelinonTaskEnvelope())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import abc


class DataStorage(metaclass=abc.ABCMeta):
    """
    Abstract Selinon storage adapter that is implemented by a user
    """
    @abc.abstractclassmethod
    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def connect(self):
        """
        Connect to a resource, if not needed, should be empty
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def is_connected(self):
        """
        :return: True if connected to a resource
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def disconnect(self):
        """
        Disconnect from a resource
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve(self, task_name, task_id):
        """
        Retrieve result stored in storage

        :param task_name: task name that result is going to be retrieved
        :param task_id: id of the task that result is going to be retrieved
        :return: task result
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def store(self, node_args, flow_name, task_name, task_id, result):
        """
        Store result stored in storage

        :param node_args: arguments that were passed to node
        :param flow_name: flow name in which task was executed
        :param task_name: task name that result is going to be stored
        :param task_id: id of the task that result is going to be stored
        :param result: result that should be stored
        :return: unique ID of stored record
        """
        raise NotImplementedError()

    def __del__(self):
        if self.is_connected():
            self.disconnect()
